==> pyvispr-2024.2-py3-none-any.whl: pyvispr/extension/function.py <==
# Copyright CNRS/Inria/UniCA
# Contributor(s): Eric Debreuve (since 2017)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from __future__ import annotations

import ast as prsr
import dataclasses as dtcl
import inspect as spct
import itertools as ittl
import types as t
import typing as h
from pathlib import Path as path_t

from logger_36 import LOGGER
from pyvispr.constant.function import NO_ANNOTATION, NO_INPUT_NAME, NO_OUTPUT_NAMES
from pyvispr.extension.module import M_F_FromPathAndName, M_F_FromPyPathAndName

signature_t = spct.Signature
parameter_t = spct.Parameter

parameter_h = tuple[str, str]
parameter_w_default_h = tuple[str, str, h.Any]
parameters_h = list[parameter_h | parameter_w_default_h]
returns_h = list[parameter_h]
documentation_h = tuple[str, str, parameters_h, returns_h]


# UNSPECIFIED_INPUT_KIND = (parameter_t.VAR_POSITIONAL, parameter_t.VAR_KEYWORD)


@dtcl.dataclass(slots=True, repr=False, eq=False)
class function_t:
    module: t.ModuleType
    path: path_t | None
    pypath: str
    #
    instance: t.FunctionType
    name: str
    imports: tuple[str, ...] | None  # Limited to what annotations need.
    header_line: int
    header_inputs: dict[str, str | tuple[str, str | None]]
    header_outputs: dict[str, str] | None
    documentation: str | None
    body_line: int
    body_end_line: int

    @classmethod
    def Dummy(
        cls,
        instance: t.FunctionType,
        expected_name: str,
        module: t.ModuleType,
        path: str | None,
        pypath: str,
        /,
    ) -> function_t:
        """"""
        header_inputs = {f"{NO_INPUT_NAME}_0": f'"{NO_ANNOTATION}"'}
        header_outputs = {NO_OUTPUT_NAMES: f'"{NO_ANNOTATION}"'}

        return cls(
            module=module,
            path=path,
            pypath=pypath,
            instance=instance,
            name=expected_name,
            imports=None,
            header_line=-1,
            header_inputs=header_inputs,
            header_outputs=header_outputs,
            documentation=None,
            body_line=-1,
            body_end_line=-1,
        )

    @classmethod
    def NewFromInstance(
        cls, instance: t.FunctionType, expected_name: str, module: t.ModuleType, /
    ) -> function_t:
        """"""
        path = getattr(module, "__file__", None)
        pypath = module.__name__

        try:
            source = spct.getsource(instance)
        except TypeError:  # See documentation.
            LOGGER.error(
                f"Cannot read source of {pypath}.{instance.__name__}; Returning dummy function"
            )
            return cls.Dummy(instance, expected_name, module, path, pypath)

        imports = []
        header_inputs: dict[str, str | tuple[str, str]] = {}

        function_node = FirstFunctionAsAstNode(source)
        if function_node is None:
            LOGGER.error(
                f"Function {pypath}.{instance.__name__} cannot be read; Returning dummy function"
            )
            return cls.Dummy(instance, expected_name, module, path, pypath)

        ast_inputs = function_node.args
        positionals, keywords, keywords_defaults = _PositionalsKeywordsAndDefaults(
            ast_inputs
        )
        _AddPositionalsOrKeywords(positionals, None, source, imports, header_inputs)
        _AddVarArgs(ast_inputs, header_inputs)
        _AddPositionalsOrKeywords(
            keywords, keywords_defaults, source, imports, header_inputs
        )
        _AddKwargs(ast_inputs, header_inputs)

        if function_node.returns is None:
            header_outputs = {NO_OUTPUT_NAMES: f'"{NO_ANNOTATION}"'}
        else:
            annotation = prsr.get_source_segment(source, function_node.returns)
            if annotation == "None":
                header_outputs = None
            else:
                header_outputs = {NO_OUTPUT_NAMES: annotation}

        source_first = prsr.get_source_segment(source, function_node.body[0])
        if source_first[0].strip() == '"':
            body_line = function_node.body[1].lineno
        else:
            body_line = function_node.body[0].lineno
        source_last = prsr.get_source_segment(source, function_node.body[-1])
        body_end_line = (
            function_node.body[-1].lineno + source_last.splitlines().__len__() - 1
        )

        return cls(
            module=module,
            path=path,
            pypath=pypath,
            instance=instance,
            name=expected_name,
            imports=tuple(imports),
            header_line=function_node.lineno,
            header_inputs=header_inputs,
            header_outputs=header_outputs,
            documentation=spct.getdoc(instance),
            body_line=body_line,
            body_end_line=body_end_line,
        )

    @classmethod
    def NewFromPath(
        cls, path: path_t | str, /, *, name: str = None
    ) -> function_t | None:
        """
        name: If None, then select first function in module.
        """
        if name is None:
            assert isinstance(path, path_t)
            with open(path) as accessor:
                function = FirstFunctionAsAstNode(accessor.read())
            if function is None:
                LOGGER.error(f"No proper function found in {path}")
                return None
            name = function.name

        if isinstance(path, path_t):
            ModuleAndFunction = M_F_FromPathAndName
        else:
            ModuleAndFunction = M_F_FromPyPathAndName
        module, function = ModuleAndFunction(path, name)

        return function_t.NewFromInstance(function, name, module)

    @property
    def has_outputs(self) -> bool:
        """"""
        return self.header_outputs is not None

    @property
    def output_names(self) -> str | None:
        """"""
        if self.has_outputs:
            # This is equivalent to ", ".join(self.header_outputs)
            return NO_OUTPUT_NAMES

        return None

    @property
    def header(self) -> str:
        """
        If the function name was set to the actual function name, then the following update might be necessary:
        import re as regx
        if "__call__" in header:
            header = regx.sub(r"\b__call__\b", expected_name, header, count=1)
        This is currently not needed because NewFromInstance uses expected_name instead.
        """
        inputs = []
        positional_s = tuple(
            (_key, _vle)
            for _key, _vle in self.header_inputs.items()
            if isinstance(_vle, str)
        )
        for name, value in positional_s:
            inputs.append(f"{name}: {value}")
        if positional_s.__len__() > 0:
            inputs.append("/")
        # Could be done more efficiently with set difference, but order would be lost.
        keywords = tuple(
            (_key, _vle)
            for _key, _vle in self.header_inputs.items()
            if isinstance(_vle, tuple)
        )
        if keywords.__len__() > 0:
            inputs.append("*")
            for name, value in keywords:
                inputs.append(f"{name}: {value[0]} = {value[1]}")

        output = f"def {self.name}(" + ", ".join(inputs) + ")"

        if self.has_outputs:
            output += f" -> {self.header_outputs[NO_OUTPUT_NAMES]}:"
        else:
            output += " -> None:"

        return output

    def ModuleWithNewHeader(self, new_header: str, /) -> str:
        """
        new_header can include documentation.
        """
        if new_header[-1] == "\n":
            new_header = new_header[:-1]

        source_lines = spct.getsource(self.module).splitlines()

        prologue = source_lines[: (self.header_line - 1)]
        if self.imports is not None:
            for import_ in self.imports:
                if import_ not in prologue:
                    prologue.append(import_)
        epilogue = source_lines[(self.body_line - 1) :]

        return "\n".join(prologue + [new_header] + epilogue + [""])


def FirstFunctionAsAstNode(source: str, /) -> prsr.FunctionDef | None:
    """"""
    output = None

    try:
        tree = prsr.parse(source)
    except Exception as exception:
        LOGGER.error(str(exception))
        return None

    for node in prsr.walk(tree):
        if isinstance(node, prsr.FunctionDef) and (node.name[0] != "_"):
            output = node
            break

    return output


def _PositionalsKeywordsAndDefaults(
    ast_inputs: prsr.arguments, /
) -> tuple[list[prsr.arg], list[prsr.arg], h.Sequence[h.Any]]:
    """"""
    if (n_defaults := ast_inputs.defaults.__len__()) > 0:
        n_wo_defaults = ast_inputs.args.__len__() - n_defaults
        positionals = ast_inputs.posonlyargs + ast_inputs.args[:n_wo_defaults]
        keywords = ast_inputs.args[n_wo_defaults:]
    else:
        positionals = ast_inputs.posonlyargs + ast_inputs.args
        keywords = []
    keywords += ast_inputs.kwonlyargs
    keywords_defaults = ast_inputs.defaults + ast_inputs.kw_defaults

    return positionals, keywords, keywords_defaults


def _AddPositionalsOrKeywords(
    ast_inputs: h.Sequence[prsr.arg],
    defaults: h.Sequence[h.Any] | None,
    source: str,
    imports: list[str],
    inputs: dict[str, str | tuple[str, str]],
    /,
) -> None:
    """"""
    has_defaults = defaults is not None
    if not has_defaults:
        defaults = ()

    for ast_input, default in ittl.zip_longest(ast_inputs, defaults, fillvalue=None):
        name = ast_input.arg
        if has_defaults:
            default = prsr.get_source_segment(source, default)

        if ast_input.annotation is None:
            annotation = f'"{NO_ANNOTATION}"'
        else:
            annotation = prsr.get_source_segment(source, ast_input.annotation)
            if "." in annotation:
                imports.append(f"import {annotation[:annotation.rfind('.')]}")

        if has_defaults:
            annotation = (annotation, default)
        inputs[name] = annotation


def _AddVarArgs(ast_inputs: prsr.arguments, inputs: dict[str, str], /) -> None:
    """"""
    if ast_inputs.vararg is not None:
        name = f"{NO_INPUT_NAME}_args"
        inputs[name] = f'"{NO_ANNOTATION}"'


def _AddKwargs(
    ast_inputs: prsr.arguments, inputs: dict[str, tuple[str, str | None]], /
) -> None:
    """"""
    if ast_inputs.kwarg is not None:
        name = f"{NO_INPUT_NAME}_kwargs"
        inputs[name] = (f'"{NO_ANNOTATION}"', None)

==> pyvispr-2024.16-py3-none-any.whl: pyvispr/extension/function.py <==
"""
Copyright CNRS/Inria/UniCA
Contributor(s): Eric Debreuve (eric.debreuve@cnrs.fr) since 2017
SEE COPYRIGHT NOTICE BELOW
"""

from __future__ import annotations

import ast as prsr
import dataclasses as dtcl
import inspect as spct
import itertools as ittl
import types as t
import typing as h
from pathlib import Path as path_t

from logger_36 import LOGGER
from pyvispr.constant.function import NO_ANNOTATION, NO_INPUT_NAME, NO_OUTPUT_NAMES
from pyvispr.extension.module import M_F_FromPathAndName, M_F_FromPyPathAndName

signature_t = spct.Signature
parameter_t = spct.Parameter

parameter_h = tuple[str, str]
parameter_w_default_h = tuple[str, str, h.Any]
parameters_h = list[parameter_h | parameter_w_default_h]
returns_h = list[parameter_h]
documentation_h = tuple[str, str, parameters_h, returns_h]


@dtcl.dataclass(slots=True, repr=False, eq=False)
class function_t:
    module: t.ModuleType
    path: path_t | None
    pypath: str
    #
    instance: t.FunctionType
    name: str
    imports: tuple[str, ...] | None  # Limited to what annotations need.
    header_line: int
    header_inputs: dict[str, str | tuple[str, str | None]]
    header_outputs: dict[str, str] | None
    documentation: str | None
    body_line: int
    body_end_line: int

    @classmethod
    def Dummy(
        cls,
        instance: t.FunctionType,
        expected_name: str,
        module: t.ModuleType,
        path: str | None,
        pypath: str,
        /,
    ) -> function_t:
        """"""
        header_inputs = {f"{NO_INPUT_NAME}_0": f'"{NO_ANNOTATION}"'}
        header_outputs = {NO_OUTPUT_NAMES: f'"{NO_ANNOTATION}"'}

        return cls(
            module=module,
            path=path,
            pypath=pypath,
            instance=instance,
            name=expected_name,
            imports=None,
            header_line=-1,
            header_inputs=header_inputs,
            header_outputs=header_outputs,
            documentation=None,
            body_line=-1,
            body_end_line=-1,
        )

    @classmethod
    def NewFromInstance(
        cls, instance: t.FunctionType, expected_name: str, module: t.ModuleType, /
    ) -> function_t:
        """"""
        path = getattr(module, "__file__", None)
        pypath = module.__name__

        try:
            source = spct.getsource(instance)
        except (OSError, TypeError):  # See documentation.
            as_str = getattr(instance, "__name__", str(instance))
            LOGGER.error(
                f"Cannot read source of {pypath}.{as_str}; Returning dummy function"
            )
            return cls.Dummy(instance, expected_name, module, path, pypath)

        imports = []
        header_inputs: dict[str, str | tuple[str, str]] = {}

        function_node = FirstFunctionAsAstNode(source)
        if function_node is None:
            LOGGER.error(
                f"Function {pypath}.{instance.__name__} cannot be read; "
                f"Returning dummy function"
            )
            return cls.Dummy(instance, expected_name, module, path, pypath)

        ast_inputs = function_node.args
        positionals, keywords, keywords_defaults = _PositionalsKeywordsAndDefaults(
            ast_inputs
        )
        _AddPositionalsOrKeywords(positionals, None, source, imports, header_inputs)
        _AddVarArgs(ast_inputs, header_inputs)
        _AddPositionalsOrKeywords(
            keywords, keywords_defaults, source, imports, header_inputs
        )
        _AddKwargs(ast_inputs, header_inputs)

        if function_node.returns is None:
            header_outputs = {NO_OUTPUT_NAMES: f'"{NO_ANNOTATION}"'}
        else:
            annotation = prsr.get_source_segment(source, function_node.returns)
            if annotation == "None":
                header_outputs = None
            else:
                header_outputs = {NO_OUTPUT_NAMES: annotation}

        source_first = prsr.get_source_segment(source, function_node.body[0])
        if source_first[0].strip() == '"':
            body_line = function_node.body[1].lineno
        else:
            body_line = function_node.body[0].lineno
        source_last = prsr.get_source_segment(source, function_node.body[-1])
        body_end_line = (
            function_node.body[-1].lineno + source_last.splitlines().__len__() - 1
        )

        return cls(
            module=module,
            path=path,
            pypath=pypath,
            instance=instance,
            name=expected_name,
            imports=tuple(imports),
            header_line=function_node.lineno,
            header_inputs=header_inputs,
            header_outputs=header_outputs,
            documentation=spct.getdoc(instance),
            body_line=body_line,
            body_end_line=body_end_line,
        )

    @classmethod
    def NewFromPath(
        cls, path: path_t | str, /, *, name: str = None
    ) -> function_t | None:
        """
        name: If None, then select first function in module.
        """
        if name is None:
            assert isinstance(path, path_t)
            with open(path) as accessor:
                function = FirstFunctionAsAstNode(accessor.read())
            if function is None:
                LOGGER.error(f"No proper function found in {path}")
                return None
            name = function.name

        if isinstance(path, path_t):
            ModuleAndFunction = M_F_FromPathAndName
        else:
            ModuleAndFunction = M_F_FromPyPathAndName
        module, function = ModuleAndFunction(path, name)

        return function_t.NewFromInstance(function, name, module)

    @property
    def has_outputs(self) -> bool:
        """"""
        return self.header_outputs is not None

    @property
    def output_names(self) -> str | None:
        """"""
        if self.has_outputs:
            # This is equivalent to ", ".join(self.header_outputs)
            return NO_OUTPUT_NAMES

        return None

    @property
    def header(self) -> str:
        """
        If the function name was set to the actual function name, then the following
        update might be necessary:
        import re as regx
        if "__call__" in header:
            header = regx.sub(r"\b__call__\b", expected_name, header, count=1)
        This is currently not needed because NewFromInstance uses expected_name instead.
        """
        inputs = []
        positional_s = tuple(
            (_key, _vle)
            for _key, _vle in self.header_inputs.items()
            if isinstance(_vle, str)
        )
        for name, value in positional_s:
            inputs.append(f"{name}: {value}")
        if positional_s.__len__() > 0:
            inputs.append("/")
        # Could be done more efficiently with set difference, but order would be lost.
        keywords = tuple(
            (_key, _vle)
            for _key, _vle in self.header_inputs.items()
            if isinstance(_vle, tuple)
        )
        if keywords.__len__() > 0:
            inputs.append("*")
            for name, value in keywords:
                inputs.append(f"{name}: {value[0]} = {value[1]}")

        output = f"def {self.name}(" + ", ".join(inputs) + ")"

        if self.has_outputs:
            output += f" -> {self.header_outputs[NO_OUTPUT_NAMES]}:"
        else:
            output += " -> None:"

        return output

    def ModuleWithNewHeader(self, new_header: str, /) -> str:
        """
        new_header can include documentation.
        """
        if new_header[-1] == "\n":
            new_header = new_header[:-1]

        source_lines = spct.getsource(self.module).splitlines()

        prologue = source_lines[: (self.header_line - 1)]
        if self.imports is not None:
            for import_ in self.imports:
                if import_ not in prologue:
                    prologue.append(import_)
        epilogue = source_lines[(self.body_line - 1) :]

        return "\n".join(prologue + [new_header] + epilogue + [""])


def FirstFunctionAsAstNode(source: str, /) -> prsr.FunctionDef | None:
    """"""
    output = None

    try:
        tree = prsr.parse(source)
    except Exception as exception:
        LOGGER.error(str(exception))
        return None

    for node in prsr.walk(tree):
        if isinstance(node, prsr.FunctionDef) and (node.name[0] != "_"):
            output = node
            break

    return output


def _PositionalsKeywordsAndDefaults(
    ast_inputs: prsr.arguments, /
) -> tuple[list[prsr.arg], list[prsr.arg], h.Sequence[h.Any]]:
    """"""
    if (n_defaults := ast_inputs.defaults.__len__()) > 0:
        n_wo_defaults = ast_inputs.args.__len__() - n_defaults
        positionals = ast_inputs.posonlyargs + ast_inputs.args[:n_wo_defaults]
        keywords = ast_inputs.args[n_wo_defaults:]
    else:
        positionals = ast_inputs.posonlyargs + ast_inputs.args
        keywords = []
    keywords += ast_inputs.kwonlyargs
    keywords_defaults = ast_inputs.defaults + ast_inputs.kw_defaults

    return positionals, keywords, keywords_defaults


def _AddPositionalsOrKeywords(
    ast_inputs: h.Sequence[prsr.arg],
    defaults: h.Sequence[h.Any] | None,
    source: str,
    imports: list[str],
    inputs: dict[str, str | tuple[str, str]],
    /,
) -> None:
    """"""
    has_defaults = defaults is not None
    if not has_defaults:
        defaults = ()

    default_source = "NO DEFAULT"
    valid_default = False
    for ast_input, default in ittl.zip_longest(ast_inputs, defaults, fillvalue=None):
        name = ast_input.arg
        if has_defaults:
            default_source = prsr.get_source_segment(source, default)
            # /!\ If the default value cannot be evaluated, e.g. numpy._NoValue whose
            # import seems complicated (see Numpy source code), it is ignored, turning
            # the argument into positional.
            try:
                _ = eval(default_source)
                valid_default = True
            except:
                valid_default = False

        if ast_input.annotation is None:
            annotation = f'"{NO_ANNOTATION}"'
        else:
            annotation = prsr.get_source_segment(source, ast_input.annotation)
            if "." in annotation:
                imports.append(f"import {annotation[:annotation.rfind('.')]}")

        if has_defaults and valid_default:
            annotation = (annotation, default_source)
        inputs[name] = annotation


def _AddVarArgs(ast_inputs: prsr.arguments, inputs: dict[str, str], /) -> None:
    """"""
    if ast_inputs.vararg is not None:
        name = f"{NO_INPUT_NAME}_args"
        inputs[name] = f'"{NO_ANNOTATION}"'


def _AddKwargs(
    ast_inputs: prsr.arguments, inputs: dict[str, tuple[str, str | None]], /
) -> None:
    """"""
    if ast_inputs.kwarg is not None:
        name = f"{NO_INPUT_NAME}_kwargs"
        inputs[name] = (f'"{NO_ANNOTATION}"', None)


"""
COPYRIGHT NOTICE

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

SEE LICENCE NOTICE: file README-LICENCE-utf8.txt at project source root.

This software is being developed by Eric Debreuve, a CNRS employee and
member of team Morpheme.
Team Morpheme is a joint team between Inria, CNRS, and UniCA.
It is hosted by the Centre Inria d'Université Côte d'Azur, Laboratory
I3S, and Laboratory iBV.

CNRS: https://www.cnrs.fr/index.php/en
Inria: https://www.inria.fr/en/
UniCA: https://univ-cotedazur.eu/
Centre Inria d'Université Côte d'Azur: https://www.inria.fr/en/centre/sophia/
I3S: https://www.i3s.unice.fr/en/
iBV: http://ibv.unice.fr/
Team Morpheme: https://team.inria.fr/morpheme/
"""
